from compgeom.pnt2d import Pnt2D
from geometry.curves.line import Line
from geometry.curves.polyline import Polyline
from geometry.curves.quadbezier import QuadBezier
from geometry.curves.cubicbezier import CubicBezier
from geometry.curves.circle import Circle
from geometry.curves.circlearc import CircleArc
import math


class CurveCollector:
    def __init__(self, _model=[]):
        self.model = _model
        self.curve = None
        self.prevPt = None
        self.tempPt = None
        self.curveType = None  # LINE, POLYLINE...

    # ---------------------------------------------------------------------
    def setModel(self, _model):
        self.model = _model

    # ---------------------------------------------------------------------
    def getModel(self):
        return self.model

    # ---------------------------------------------------------------------
    def setCurveType(self, _type):
        self.curveType = _type

    # ---------------------------------------------------------------------
    def getCurveType(self):
        return self.curveType

    # ---------------------------------------------------------------------
    def startCollection(self):
        if self.curve is not None:
            self.curve = None

    # ---------------------------------------------------------------------
        if self.curveType == 'LINE':
            self.curve = Line()
        elif self.curveType == 'POLYLINE':
            self.curve = Polyline()
        elif self.curveType == 'QUADBEZIER':
            self.curve = QuadBezier()
        elif self.curveType == 'CUBICBEZIER':
            self.curve = CubicBezier()
        elif self.curveType == 'CIRCLE':
            self.curve = Circle()
        elif self.curveType == 'CIRCLEARC':
            self.curve = CircleArc()

    # ---------------------------------------------------------------------
    def endCollection(self, _tol):
        self.model.insertCurve(self.curve, _tol)
        self.curve = None
        self.tempPt = None

    # ---------------------------------------------------------------------
    def isActive(self):
        if self.curve is not None:
            return True
        return False

    # ---------------------------------------------------------------------
    def isCollecting(self):
        if self.curve is not None:
            if self.curve.getNumberOfCtrlPoints() > 0:
                return True
        return False

    # ---------------------------------------------------------------------
    def hasFinished(self):
        if self.curve is not None:
            if self.curve.isPossible():
                return True
        return False

    # ---------------------------------------------------------------------
    def isUnlimited(self):
        if self.curve is not None:
            if self.curve.isUnlimited():
                return True
        return False

    # ---------------------------------------------------------------------
    def insertCtrlPoint(self, _x, _y, _tol):
        if self.isCollecting():
            if self.prevPt is not None:
                if((abs(_x - self.prevPt.getX()) <= _tol) and
                   (abs(_y - self.prevPt.getY()) <= _tol)):
                    return 0

        self.curve.addCtrlPoint(_x, _y)
        if self.prevPt is None:
            self.prevPt = Pnt2D(_x, _y)
        else:
            self.prevPt.setCoords(_x, _y)
        self.tempPt = None
        return 1

    # ---------------------------------------------------------------------
    def addTempCtrlPoint(self, _x, _y):
        if self.tempPt is None:
            self.tempPt = Pnt2D(_x, _y)
        else:
            self.tempPt.setCoords(_x, _y)
        return 1

    # ---------------------------------------------------------------------
    def getCurveCollected(self):
        return self.curve

    # ---------------------------------------------------------------------
    def getDrawPoints(self):
        if self.curve is None:
            return []
        if self.tempPt is None:
            return []
        return self.curve.getEquivPolylineCollecting(self.tempPt)

    # ---------------------------------------------------------------------
    def getCtrlPoints(self):
        if self.curve is None:
            ctrlPts = []
        else:
            ctrlPts = self.curve.getCtrlPoints()
        if ctrlPts == []:
            pts = []
        else:
            pts = ctrlPts.copy()
        if self.tempPt is not None:
            pts.append(self.tempPt)
        return pts

    # ---------------------------------------------------------------------
    def getBoundBox(self):
        if self.curve is None:
            return
        return self.curve.getBoundBox()

    # ---------------------------------------------------------------------
    def reset(self):
        self.curve = None

    # ---------------------------------------------------------------------
    def kill(self):
        if self.curve is not None:
            del self.curve
        del self

    # ---------------------------------------------------------------------
    def snapToCurrentCurve(self, _x, _y, _tol):

        if self.curve is None:
            return False, _x, _y

        if not self.curve.isUnlimited():
            return False, _x, _y

        pts = self.curve.getCtrlPoints()

        if len(pts) < 3:
            return False, _x, _y

        snap = False
        dmin = _tol

        for i in range(0, len(pts)):
            pt_x = pts[i].getX()
            pt_y = pts[i].getY()
            d = math.sqrt((_x - pt_x) * (_x - pt_x) +
                          (_y - pt_y) * (_y - pt_y))

            if d < dmin:
                xmin = pt_x
                ymin = pt_y
                dmin = d
                snap = True

        if snap:
            return True, xmin, ymin
        else:
            status, clstPt, dist, t, tangVec = self.curve.closestPoint(_x, _y)
            if not status:
                return False, _x, _y
            if dist < dmin:
                dmin = dist
                xmin = clstPt.getX()
                ymin = clstPt.getY()
                return True, xmin, ymin
            else:
                return False, _x, _y
