from mesh.meshgenerator import MeshGenerator
from compgeom.pnt2d import Pnt2D
from compgeom.compgeom import CompGeom
import triangle as tr
import math


# This class generates meshes by the method of constrained Delaunay triangulation.
# It uses Jonathan Richard Shewchuk’s Triangle - "A Two-Dimensional Quality Mesh
# Generator and Delaunay Triangulator" (http://www.cs.cmu.edu/~quake/triangle.html).
# Copyright: Jonathan Richard Shewchuk (http://people.eecs.berkeley.edu/~jrs/).
# Triangle is implemented in C and this class uses a Python wrapper 
# (https://rufat.be/triangle) around the C implementation developed by The Block
# Research Group (BRG) at the Institute of Technology in Architecture at
# ETH Zürich (https://block.arch.ethz.ch/brg/about).
#
# This use of Triangle is limited to one region with only an external loop
# (no holes).
#
# Triangle generates Delaunay triangulations constrained to the given list of
# loop points. However, to make Triangle's use more robust, the only specified
# parameter is a maximum triangle area constraint, which is calculated as the
# area of an equilateral triangle whose side is equal to the maximum segment
# size of the input loop. Since this is the only parameter specified for
# Triangle, in the case of concave regions, it generates triangles outside the
# given loop and inside its convex hull. Therefore, these external triangles
# and their vertices' points are deleted in the post-processing stage.
#
# In addition, Triangle may generate new triangulation points on the
# external loop. No fix is made for this.
#
# Installation: pip install triangle
#

class MeshDelaunay(MeshGenerator):

    def __init__(self):
        pass

    # ---------------------------------------------------------------------
    @staticmethod
    def pickLoop(loopPts, _pt, _tol):
        nLoopPts = len(loopPts)
        for i in range(0, nLoopPts):
            if CompGeom.pickLine(loopPts[i], loopPts[(i + 1)%nLoopPts], _pt, _tol):
                return True
        return False

    # ---------------------------------------------------------------------
    def setLoops(self, _loops):

        # Check for just one loop
        if len(_loops) != 1:
            return False

        return True

    # ---------------------------------------------------------------------
    def generateMesh(self, _bdryPnts):

        inputPts = []
        inputPts.append([_bdryPnts[0].getX(), _bdryPnts[0].getY()])
        maxL = Pnt2D.euclidiandistance(_bdryPnts[0], _bdryPnts[1])
        nBdryPts = len(_bdryPnts)
        for i in range(1, nBdryPts):
            inputPts.append([_bdryPnts[i].getX(), _bdryPnts[i].getY()])
            sizeL = Pnt2D.euclidiandistance(_bdryPnts[i], _bdryPnts[(i + 1)%nBdryPts])
            if sizeL > maxL:
                maxL = sizeL

        maxArea = maxL*maxL*math.sqrt(3.0)/4.0; # equilateral triangle area
        att = "a" + str(maxArea)
        t = tr.triangulate({'vertices': inputPts}, att)
        ptsRaw = t['vertices'].tolist()
        connRaw = t['triangles'].tolist()

        # Create a list of reduced triangulation points considering only
        # points inside the given loop, including the loop points.
        # In addition, create a list with the indices of the points in the
        # reduced triangulation. An outside point will have an invalid (-1)
        # index in this list.
        tol = maxL * 0.001
        pts = []
        newPtIds = []
        newId = 0
        for i in range(0, nBdryPts):
            pt = Pnt2D(ptsRaw[i][0], ptsRaw[i][1])
            pts.append(pt)
            newPtIds.append(newId)
            newId += 1
        for i in range(nBdryPts, len(ptsRaw)):
            pt = Pnt2D(ptsRaw[i][0], ptsRaw[i][1])
            if MeshDelaunay.pickLoop(_bdryPnts, pt, tol):
                pts.append(pt)
                newPtIds.append(newId)
                newId += 1
            elif CompGeom.isPointInPolygon(_bdryPnts, pt):
                pts.append(pt)
                newPtIds.append(newId)
                newId += 1
            else:
                newPtIds.append(-1)

        # Create a list of reduced triangulation triangles considering only
        # triangles inside the loop.
        conn = []
        for i in range(0, len(connRaw)):
            inside = True
            nTriangleBdryPts = 0
            xCenter = 0.0
            yCenter = 0.0
            for j in range(0, 3):
                if connRaw[i][j] < nBdryPts:
                    nTriangleBdryPts += 1
                if newPtIds[connRaw[i][j]] == -1:
                    inside = False
                xCenter += ptsRaw[connRaw[i][j]][0]
                yCenter += ptsRaw[connRaw[i][j]][1]
            xCenter /= 3.0
            yCenter /= 3.0
            if inside and (nTriangleBdryPts == 3):
                pt = Pnt2D(xCenter, yCenter)
                if not CompGeom.isPointInPolygon(_bdryPnts, pt):
                    inside = False
            if inside:
                v0 = newPtIds[connRaw[i][0]]
                v1 = newPtIds[connRaw[i][1]]
                v2 = newPtIds[connRaw[i][2]]
                conn.append([v0, v1, v2])
            
        return True, pts, conn
